import * as common from '/pages/src/common.mjs';
import * as chart from './fitness-chart.mjs';
import { initAthleteState, calculatePowerCurve, initMod } from './server-requests.mjs';
import { AthleteData } from './athleteData.mjs';

const UPDATE_INTERVAL = 5000; // How often to update the window's data, in milliseconds

common.settingsStore.setDefault({
  updateWatchingCp: false,
  updateNearbyCp: false, // setting removed; trying to get data for everyone in an event blows up ZP rate limiting.
  showChartCP: true,
  showChartSmoothPower: true,
  showEstimatedPowersTable: false,
});

function initHtmlElements() {
  return {
    powersTable: document.querySelector('#powers-table'),

    zFtpWatts: document.querySelector('#zftp-watts'),
    zFtpMaxWatts: document.querySelector('#zftp-max-watts'),
    zFtpWattsLimit: document.querySelector('#zftp-watts-limit'),

    zFtpWkg: document.querySelector('#zftp-wkg'),
    zFtpMaxWkg: document.querySelector('#zftp-max-wkg'),
    zFtpWkgLimit: document.querySelector('#zftp-wkg-limit'),

    zMapWatts: document.querySelector('#zmap-watts'),
    // zMapRealWatts: document.querySelector('#zmap-real-watts'),
    zMapMaxWatts: document.querySelector('#zmap-max-watts'),
    zMapWattsLimit: document.querySelector('#zmap-watts-limit'),

    zMapWkg: document.querySelector('#zmap-wkg'),
    // zMapRealWkg: document.querySelector('#zmap-real-wkg'),
    zMapMaxWkg: document.querySelector('#zmap-max-wkg'),
    zMapWkgLimit: document.querySelector('#zmap-wkg-limit'),

    peakPowerChart: document.querySelector('#peak-power-chart'),
    categoryLetter: document.querySelector('#category-letter'),
    messages: document.querySelector('#messages'),
    maxTimeForPower: document.querySelector('#max-time-for-power .stats-value'),
    wPrimeBal: document.querySelector('#wprime-bal .stats-value'),
  };
}

function render(state, athleteData, fitness) {
  if (state.settings.showEstimatedPowersTable) {
    document.body.classList.add('showEstimatedPowersTable');
  } else {
    document.body.classList.remove('showEstimatedPowersTable');
  }

  state.elements.zFtpWatts.textContent = fitness?.zFtp?.watts || '-';
  state.elements.zFtpMaxWatts.textContent = fitness?.zFtpMax?.watts || '-';
  state.elements.zFtpWattsLimit.textContent = athleteData?.category?.zFtpWattsLimit || '-';

  state.elements.zFtpWkg.textContent = fitness?.zFtp?.wkg || '-';
  state.elements.zFtpMaxWkg.textContent = fitness?.zFtpMax?.wkg || '-';
  state.elements.zFtpWkgLimit.textContent = athleteData?.category?.zFtpWkgLimit || '-';

  state.elements.zMapWatts.textContent = fitness?.zMap?.watts || '-';
  // state.elements.zMapRealWatts.textContent = fitness?.zMapReal?.watts || '-';
  state.elements.zMapMaxWatts.textContent = fitness?.zMapMax?.watts || '-';
  state.elements.zMapWattsLimit.textContent = athleteData?.category?.zMapWattsLimit || '-';

  state.elements.zMapWkg.textContent = fitness?.zMap?.wkg || '-';
  // state.elements.zMapRealWkg.textContent = fitness?.zMapReal?.wkg || '-';
  state.elements.zMapMaxWkg.textContent = fitness?.zMapMax?.wkg || '-';
  state.elements.zMapWkgLimit.textContent = athleteData?.category?.zMapWkgLimit || '-';

  state.elements.categoryLetter.textContent = athleteData?.category?.letter || '';

  if(athleteData?.wPrimeBal) {
    state.elements.wPrimeBal.textContent = `${athleteData.pctWPrime}%`;
    state.elements.wPrimeBal.style.color = calculateColorFromGreenToRed(100, 0, athleteData.pctWPrime);
  } else {
    state.elements.wPrimeBal.textContent = '-';
    state.elements.wPrimeBal.style.color = 'white';
  }

  while (state.elements.messages.firstChild) {
    state.elements.messages.removeChild(state.elements.messages.firstChild);
  }
  const allMessages = [
    ...state.messages || [],
    ...athleteData?.messages || [],
    ...fitness?.messages || []
  ];
  if(allMessages.length) {
    allMessages.forEach(m => {
      const li = document.createElement('li');
      li.textContent = m;
      state.elements.messages.appendChild(li);
    });
    state.elements.messages.classList.remove('hidden');
  } else {
    state.elements.messages.classList.add('hidden');
  }

  const zFtpRegression = [];
  const zFtpPoint = [];
  if(fitness?.zFtp?.regression?.intercept) {
    zFtpRegression.push([60 * state.const.ZFTP_LOWER_MINUTE, fitness.zFtp.regression.intercept + fitness.zFtp.regression.slope * Math.log(60 * state.const.ZFTP_LOWER_MINUTE)]);
    // zFtpRegression.push([fitness.zFtp.time, fitness.zFtp.watts]);
    zFtpRegression.push([60 * state.const.ZFTP_UPPER_MINUTE, fitness.zFtp.regression.intercept + fitness.zFtp.regression.slope * Math.log(60 * state.const.ZFTP_UPPER_MINUTE)]);
    zFtpPoint.push([fitness.zFtp.time, fitness.zFtp.watts]);
  }

  const zFtpMaxRegression = [];
  const zFtpMaxPoint = [];
  if(fitness?.zFtpMax?.complete && fitness?.zFtpMax?.regression?.intercept) {
    zFtpMaxRegression.push([60 * state.const.ZFTP_LOWER_MINUTE, fitness.zFtpMax.regression.intercept + fitness.zFtpMax.regression.slope * Math.log(60 * state.const.ZFTP_LOWER_MINUTE)]);
    // zFtpMaxRegression.push([fitness.zFtpMax.time, fitness.zFtpMax.watts]);
    zFtpMaxRegression.push([60 * state.const.ZFTP_UPPER_MINUTE, fitness.zFtpMax.regression.intercept + fitness.zFtpMax.regression.slope * Math.log(60 * state.const.ZFTP_UPPER_MINUTE)]);
    zFtpMaxPoint.push([fitness.zFtpMax.time, fitness.zFtpMax.watts]);
  }

  const zMapProjection = [];
  const zMapPoint = [];
  if(fitness?.zMap?.projection?.intercept) {
    zMapProjection.push([60 * state.const.ZMAP_LOWER_MINUTE, fitness.zMap.projection.intercept + fitness.zMap.projection.slope * Math.log(60 * state.const.ZMAP_LOWER_MINUTE)]);
    zMapProjection.push([fitness.zMap.time, fitness.zMap.watts]);
    zMapProjection.push([60 * state.const.ZMAP_UPPER_MINUTE, fitness.zMap.projection.intercept + fitness.zMap.projection.slope * Math.log(60 * state.const.ZMAP_UPPER_MINUTE)]);
    zMapPoint.push([fitness.zMap.time, fitness.zMap.watts]);
  }

  const zMapMaxProjection = [];
  const zMapMaxPoint = [];
  if(fitness?.zMapMax?.complete && fitness?.zMapMax?.regression?.intercept) {
    zMapMaxProjection.push([60 * state.const.ZMAP_LOWER_MINUTE, fitness.zMapMax.projection.intercept + fitness.zMapMax.projection.slope * Math.log(60 * state.const.ZMAP_LOWER_MINUTE)]);
    zMapMaxProjection.push([fitness.zMapMax.time, fitness.zMapMax.watts]);
    zMapMaxProjection.push([60 * state.const.ZMAP_UPPER_MINUTE, fitness.zMapMax.projection.intercept + fitness.zMapMax.projection.slope * Math.log(60 * state.const.ZMAP_UPPER_MINUTE)]);
    zMapMaxPoint.push([fitness.zMapMax.time, fitness.zMapMax.watts]);
  }

  const current = [];
  if(state.settings.showChartSmoothPower && athleteData?.sauceData && state?.periods?.times) {
    const minGraphedTime = 60;
    Object.entries(athleteData.sauceData.stats.power.smooth)
      .filter(([t,]) => t >= minGraphedTime)
      .forEach(([t, pow]) => current.push([parseInt(t), pow]));
  }
  
  const seriesData = {
    max90day: (state?.periods?.times || []).map(t => [t, athleteData?.powerProfile?.cpBestEfforts?.pointsWatts?.[t.toString()]?.value || 0]).filter(d => d[1] > 0),
    max: (state?.periods?.times || []).map(t => [t, athleteData?.maxPower[t.toString()]]).filter(d => d[1] > 0),
    current,
    zFtpRegression,
    zFtpPoint,
    zFtpMaxRegression,
    zFtpMaxPoint,
    zMapProjection,
    zMapPoint,
    zMapMaxProjection,
    zMapMaxPoint,
  };
  if(state.peakPowerChart) {
    state.peakPowerChart.updateData(seriesData, state.settings.showChartCP ? athleteData?.criticalPower : undefined);
  }
}

function renderMaxTimeForPower(state, maxTimeForPower) {
  state.elements.maxTimeForPower.textContent = maxTimeForPower ? `${maxTimeForPower}` : '-';
}

async function updateData(state) {
  if (state.initialized) {
    if (state.watchingId && state.athleteData[state.watchingId]) {
      console.log(`Power Curve Mod : updateData for ${state.watchingId}`);
      const athleteData = state.athleteData[state.watchingId];
      const startTime = athleteData.latestRollTime;
      const streams = await common.rpc.getAthleteStreams(state.watchingId, {startTime});
      if (streams) {
        athleteData.importStreams(streams.time, streams.power);

        // if (athleteData.criticalPower, athleteData.wPrime) {
        //   athleteData.wPrimeBal = sauce.power.calcWPrimeBalIntegralStatic(streams.power, streams.time, athleteData.criticalPower, athleteData.wPrime);
        // }
      }

      try {
        const fitness = await calculatePowerCurve(state, athleteData);
        render(state, athleteData, fitness);
      } catch(e) {
        state.messages = ['Error getting power data.'];
        console.error(state.messages, e);
        render(state, athleteData, undefined);
      }
    } else {
      state.messages = ['No athlete being watched.'];
      render(state, undefined, undefined);  
    }
  } else {
    state.messages = ['Waiting for game connection.'];
    render(state, undefined, undefined);
  }
}

// this function taken from Sauce pages/src/nearby.mjs
function makeLazyGetter(cb) {
  const getting = {};
  const cache = new Map();

  return function getter(key) {
      if (!cache.has(key)) {
          if (!getting[key]) {
              getting[key] = cb(key).then(value => {
                  cache.set(key, value || null);
                  if (!value) {
                      // Allow retry, especially for event data which is wonky
                      setTimeout(() => cache.delete(key), 10000);
                  }
                  delete getting[key];
              });
          }
          return;
      } else {
          return cache.get(key);
      }
  };
}
const lazyGetSubgroup = makeLazyGetter(id => common.rpc.getEventSubgroup(id));

async function init() {
  console.log('Power Curve Mod : init');

  const state = {
    initialized: false,
    athleteData: {},
    elements: initHtmlElements(),
    messages: [],
    settings: common.settingsStore.get(),
  };

  common.initInteractionListeners();

  common.settingsStore.addEventListener('changed', ev => {
    const changed = ev.data.changed;
    console.log('Power Curve Mod : settings changed', ev);
    state.settings = common.settingsStore.get();
    if (changed.has('updateWatchingCp') && state.settings.updateWatchingCp) {
      const ad = state.athleteData[state.watchingId];
      if (ad.wPrime && ad.criticalPower) {
        console.log(`Power Curve Mod : update watching ${state.watchingId} CP ${ad.criticalPower} W' ${ad.wPrime}`);
        common.rpc.updateAthlete(state.watchingId, {wPrime: ad.wPrime, cp: ad.criticalPower});
      }
    }
  });

  function updateLoop() {
    updateData(state).finally(() => setTimeout(updateLoop, UPDATE_INTERVAL));
  }
  updateLoop();

  // Called when our own data changes.
  // 'athlete/watching' is also triggered, if we are watching ourself.
  common.subscribe('athlete/self', async self => {
    if(!state.selfId) {
      state.selfId = self.athleteId;
      const initModResponse = await initMod(state);
      Object.assign(state, initModResponse.state || {});
      console.log('Power Curve Mod : initMod state', state);

      state.peakPowerChart = chart.createPeakPowerChart(state.elements.peakPowerChart, {
        xMin: state.periods.times[0],
        xMax: 60 * 60,
      });
      document.body.onresize = () => {
        state.peakPowerChart.resize();
      }
    
      state.initialized = true;
    }
  });

  // Called when the watched athlete changes
  common.subscribe('athlete/watching', async watching => {
    if (!state.selfId || !state.initialized) {
      return;
    }

    if(state.watchingId !== watching.athleteId) {
      console.log('Power Curve Mod : now watching', watching.athleteId);
      state.watchingId = watching.athleteId;

      // start tracking the watched athlete's data
      if(!state.athleteData[state.watchingId]) {
        console.log('Power Curve Mod : start watching', state.watchingId);
        state.messages = [];

        const ad = new AthleteData(state.periods.times);
        state.athleteData[state.watchingId] = ad;
        ad.sauceData = watching;

        // Sauce might already have some power numbers.
        Object.keys(watching.stats.power.peaks).forEach(period => {
          ad.maxPower[period.toString()] = watching.stats.power.peaks[period].avg;
        });
        await loadWatchedAthletePowerProfile(state, ad);

        // update Sauce's internal DB of CP and W'
        if (state.settings.updateWatchingCp && ad.wPrime && ad.criticalPower) {
          console.log(`Power Curve Mod : update watching ${x.athlete.id} CP ${ad.criticalPower} W' ${ad.wPrime}`);
          common.rpc.updateAthlete(state.watchingId, {wPrime: ad.wPrime, cp: ad.criticalPower});
        }
      }

      updateData(state); // update and rerender since watched rider changed.
    }
    const ad = state.athleteData[state.watchingId];
    const prevEventStartPending = ad.sauceData.eventStartPending;
    ad.sauceData = watching;

    // try to determine when an event starts and reset our power data.
    if(prevEventStartPending && !ad.sauceData.eventStartPending) {
      console.log('Power Curve Mod : event started, reset power data');

      // actually not sure what to do here, since we get player streams from the start of the ride.
      // would have to set the startTime to the start of the event maybe?
    }

    renderMaxTimeForPower(state, calculateMaxTimeForPower(state, watching.state.power));
  });

  common.subscribe('nearby', async nearby => {
    // if the watched rider is in an event, load the event data.
    const watching = nearby.find(x => x.watching);
    const sgid = watching && watching.state.eventSubgroupId;
    if (!sgid) {
      return;
    }

    // filter only nearby athletes in the same event subgroup.
    nearby = nearby.filter(x => x.state.eventSubgroupId === sgid);
    
    // load the event subgroup data.
    const eventGroup = lazyGetSubgroup(sgid);
    if (!state.settings.updateNearbyCp || !eventGroup || eventGroup.event.eventType !== 'RACE' || eventGroup.event.eventType !== 'TEAM_TIME_TRIAL') {
      return;
    }

    // loop over all nearby athletes and update their CP and W'.
    for (const x of nearby) {
      try {
        // start tracking the watched athlete's data
        if(!state.athleteData[x.athlete.id]) {

          const ad = new AthleteData(state.periods.times);
          state.athleteData[x.athlete.id] = ad;
          ad.sauceData = watching;
          await loadWatchedAthletePowerProfile(state, ad);

          // update Sauce's internal DB of CP and W'
          if (ad.wPrime && ad.criticalPower) {
            console.log(`Power Curve Mod : update nearby ${x.athlete.id} CP ${ad.criticalPower} W' ${ad.wPrime}`);
            common.rpc.updateAthlete(athlete.athleteId, {wPrime: ad.wPrime, cp: ad.criticalPower});
          }
        }
      } catch (e) {
        console.error('Power Curve Mod : Error updating nearby athlete', e, x);
      }
    }
  });
}

// async function handleAthleteUpdate(watching) {
//   const athleteData = state.athleteData[watching.athleteId];
//   athleteData.watching = watching;
// }

async function loadWatchedAthletePowerProfile(state, athleteData) {
  athleteData.powerProfile = undefined;
  try {
    if (athleteData.sauceData.self && common.rpc.getPowerProfile) {
      console.log('Power Curve Mod : getPowerProfile')
      athleteData.powerProfile = await common.rpc.getPowerProfile();
      athleteData.category.letter = athleteData.powerProfile.displayFemaleCategory ? `w${athleteData.powerProfile.categoryWomen}` : athleteData.powerProfile.category;
      athleteData.known = athleteData.powerProfile ? {
        zftp: athleteData.powerProfile.zftp,
        zmap: athleteData.powerProfile.zmap,
      } : undefined;
    }
  } catch (e) {
    // intentionally ignoring errors.
    // getPowerProfile was added in Sauce 1.0.0. It may not exist yet for some users.
  }

  // get initial athlete state from server
  try {
    const serverState = await initAthleteState(state, athleteData);
    athleteData.assignState(serverState);
  } catch (e) {
    state.messages = ['Error initializing athlete state'];
    console.error(state.messages, e);
  }

  const categoryInfo = athleteData.category.letter ? state.categoryLimits[athleteData.category.letter] : undefined;
  if(categoryInfo) {
    athleteData.category.zFtpWattsLimit = Math.round(categoryInfo.zFtpWkg * athleteData.sauceData.athlete.weight);
    athleteData.category.zFtpWkgLimit = categoryInfo.zFtpWkg;
    athleteData.category.zMapWattsLimit = Math.round(categoryInfo.zMapWkg * athleteData.sauceData.athlete.weight);
    athleteData.category.zMapWkgLimit = categoryInfo.zMapWkg;
  }
}

function calculateMaxTimeForPower(state, power) {
  const ad = state.athleteData[state.watchingId];
  const pointsWatts = ad?.powerProfile?.cpBestEfforts?.pointsWatts;

  // max power data is required to calculate this.
  if(!pointsWatts) {
    return undefined;
  }

  if(power <= Math.max(ad.maxPower[(60 * 60).toString()] || 0, pointsWatts[(60 * 60).toString()].value || 0)) {
    return '> 1h';
  }

  for(let i = 0; i < state.periods.times.length; i++) {
    const time = state.periods.times[i];
    if(Math.max(ad.maxPower[time.toString()] || 0, pointsWatts[time.toString()].value || 0) < power) {
      const minutes = Math.floor(time / 60);
      const seconds = time % 60;
      if(minutes > 0) {
        return `${minutes}m${seconds < 10 ? '0' : ''}${seconds}s`;
      } else {
        return `${seconds}s`;
      }
    }
  }
}

function calculateColorFromGreenToRed(high, low, value) {
  const ratio = (value - high) / (low - high);
  const hue = (1 - ratio) * 120;
  return `hsl(${hue}, 100%, 50%)`;
}

export async function settingsMain() {
  common.initInteractionListeners();
  await common.initSettingsForm('form')();
}

export async function main() {
  try {
    await init();
  } catch (e) {
    console.error('Power Curve Mod Error', e);
    if(common.showError) {
      common.showError(e);
    }
  }
}
