import { locale } from '/shared/sauce/index.mjs';
import * as common from '/pages/src/common.mjs';
import * as echarts from '/pages/deps/src/echarts.mjs';
import { cssColor } from '/pages/src/echarts-sauce-theme.mjs';

const GAUGE_MAX = 8; // w/kg

const state = {
  elements: {},
  inGroup: false,
  avgLeadWkg: 0,
  groupAvgWkg: 0,
  watchedProjectedWkg: 0,
  watchedActualWkg: 0,
  watchedMinWkg: 0,
};

const settingsStore = new common.SettingsStore('zwiftools-group-power-v1');
var settings = settingsStore.get(null, {
  refreshInterval: 1,
  dataSmoothing: 0,
  showAverage: false,
  showMax: false,
  currentLap: false,
  boringMode: false,
  gaugeTransparency: 20,
  solidBackground: false,
  backgroundColor: '#000000',
  color: '#33dd33',
  idealRangeColor: '#556655',
  exceedingRangeColor: '#665555',
  gapAhead: 2,
  gapBehind: 2,
});
common.themeInit(settingsStore);
common.localeInit(settingsStore);

function colorAlpha(color, alpha) {
  if (color.length <= 5) {
    return color.slice(0, 4) + alpha[0];
  } else {
    return color.slice(0, 7) + alpha.padStart(2, alpha[0]);
  }
}

function render() {
  state.elements.watchingPower.textContent = state.watchedActualWkg.toFixed(1);
  if (state.inGroup) {
    state.elements.groupPower.textContent = state.watchedMinWkg.toFixed(1);
    state.elements.leadPower.textContent = isNaN(state.avgLeadWkg) ? '-' : state.avgLeadWkg.toFixed(1);
  } else {
    state.elements.groupPower.textContent = '-';
    state.elements.leadPower.textContent = '-';
  }

  // make the minimum line thickness 0.1 just to make it actually visible.
  if (state.avgLeadWkg - state.watchedMinWkg < 0.1) {
    state.watchedMinWkg = state.avgLeadWkg - 0.1;
  }

  const c1 = state.watchedMinWkg;
  const c2 = state.watchedProjectedWkg > state.watchedMinWkg ? state.watchedProjectedWkg : state.watchedMinWkg;
  const c3 = state.avgLeadWkg;

  const colorBands = state.inGroup ? [
    [c1 / GAUGE_MAX, '#000000'],
    [c2 / GAUGE_MAX, settings.idealRangeColor || '#556655'],
    [c3 / GAUGE_MAX, settings.exceedingRangeColor || '#665555'],
    [1, '#000000'],
  ] : [
    [1, '#000000'],
  ];

  const series = {
    axisLine: {
      lineStyle: {
        color: colorBands,
      },
    },
  };
  series.data = [{
    value: state.watchedActualWkg.toFixed(1),
  // }, {
  //   value: state.watchedProjectedWkg.toFixed(1),
  }];
  state.gauge?.setOption({
    series: [series],
  });
}

function initHtmlElements() {
  return {
    content: document.querySelector('#content'),
    gauge: document.querySelector('#gauge'),
    groupPower: document.querySelector('#group'),
    watchingPower: document.querySelector('#watching'),
    leadPower: document.querySelector('#lead'),
  };
}

function initGauge() {
  state.gauge = echarts.init(state.elements.gauge, 'sauce', { renderer: 'svg' });
  const relSize = Math.min(state.elements.content.clientHeight * 1.20, state.elements.content.clientWidth) / 600;
  state.gauge.setOption({
    animationDurationUpdate: Math.max(200, Math.min(settings.refreshInterval * 1000, 1000)),
    animationEasingUpdate: 'linear',
    color: [
      settings.color,
      '#555555',
    ],
    graphic: [{
      elements: [{
        left: 'center',
        top: 'middle',
        type: 'circle',
        shape: {
          r: 270 * relSize,
        },
        style: {
          shadowColor: cssColor('fg', 1, 2 / 3),
          shadowBlur: 5 * relSize,
          fill: {
            type: 'linear',
            x: 0,
            y: 0,
            x2: 0,
            y2: 1,
            colorStops: [{
              offset: 0,
              color: '#000',
            }, {
              offset: 0.5,
              color: colorAlpha(settings.color, 'f'),
            }, {
              offset: 0.75,
              color: settings.color,
            }, {
              offset: 0.75,
              color: '#0000'
            }, {
              offset: 1,
              color: '#0000'
            }],
          },
          lineWidth: 0,
          opacity: 1 - (settings.gaugeTransparency / 100),
        }
      }]
    }],
    series: [{
      radius: '90%', // fill space
      splitNumber: GAUGE_MAX * 2,
      name: 'Group Power',
      type: 'gauge',
      min: 0,
      max: GAUGE_MAX,
      startAngle: 210,
      endAngle: 330,
      progress: {
        show: false,
        // width: 60 * relSize,
        // itemStyle: {
        //   color: 'transparent',
        // },
      },
      axisLine: {
        lineStyle: {
          color: [[1, cssColor('fg', 1, 0.2)]],
          width: 60 * relSize,
          shadowColor: cssColor('fg', 1, 0.5),
          shadowBlur: 8 * relSize,
        },
      },
      axisTick: {
        show: false,
      },
      splitLine: {
        show: true,
        distance: 10 * relSize,
        length: 10 * relSize,
        lineStyle: {
          width: 3 * relSize,
        }
      },
      axisLabel: {
        distance: 70 * relSize,
        fontSize: 20 * relSize,
        // formatter: config.getLabel,
        textShadowColor: cssColor('fg', 1, 0.5),
        textShadowBlur: 1 * relSize,
      },
      pointer: settings.boringMode ? {
        // NOTE: Important that all are set so it's not an update
        icon: null,
        width: 6 * relSize,
        length: 180 * relSize,
        offsetCenter: [0, 0],
        itemStyle: {
          // color: settings.color,
          opacity: 0.9,
          borderColor: cssColor('fg', 0, 0.9),
          borderWidth: 2 * relSize,
          shadowColor: cssColor('fg', 1, 0.4),
          shadowBlur: 4 * relSize,
        },
      } : {
        width: 70 * relSize,
        length: 180 * relSize,
        icon: 'image://../../../pages/images/logo_vert_120x320.png',
        offsetCenter: [0, '10%'],
      },
      anchor: settings.boringMode ? {
        showAbove: true,
        show: true,
        size: 25 * relSize,
        itemStyle: {
          color: settings.color,
          borderColor: cssColor('fg', 0, 0.9),
          borderWidth: 2 * relSize,
          shadowColor: cssColor('fg', 1, 0.5),
          shadowBlur: 4 * relSize,
        }
      } : { show: false },
      detail: {
        valueAnimation: true,
        formatter: x => `{value|${locale.human.wkg(x)}}\n{unit|wkg}`,
        textShadowColor: cssColor('fg', 1, 0.4),
        textShadowBlur: 1 * relSize,
        offsetCenter: [0, '32%'],
        rich: {
          value: {
            color: cssColor('fg'),
            fontSize: 80 * relSize,
            fontWeight: 'bold',
            lineHeight: 70 * relSize,
          },
          unit: {
            fontSize: 18 * relSize,
            color: cssColor('fg', 0, 0.88),
            lineHeight: 16 * relSize,
          }
        }
      },
    }],
  });
}

function setBackground() {
  const doc = document.documentElement;
  const {solidBackground, backgroundColor} = settings;
  doc.classList.toggle('solid-background', solidBackground);
  if (solidBackground) {
      doc.style.setProperty('--background-color', backgroundColor);
  } else {
      doc.style.removeProperty('--background-color');
  }
}

async function init() {
  console.log('ZwifTools Mod : Group Pace : init');
  
  setBackground();

  common.initInteractionListeners();
  state.elements = initHtmlElements();
  initGauge();

  const renderer = new common.Renderer(content, {fps: 1 / settings.refreshInterval});
  renderer.addCallback(render);

  settingsStore.addEventListener('set', ev => {
    const key = ev.data.key;
    setBackground();
    renderer.fps = 1 / settings.refreshInterval;
    initGauge();
    state.gauge.setOption({series: [{animation: false}]});
    // clearTimeout(reanimateTimeout);
    renderer.render({force: true});
    // setTimeout(() => state.gauge.setOption({series: [{animation: true}]}), 400);
  });

  common.subscribe('nearby', async nearby => {
    const watching = nearby.find(x => x.watching);
    const filtered = nearby
      .filter(x => x.athlete && x.state.power > 0 && x.gap <= settings.gapBehind && x.gap >= (settings.gapAhead * -1));
    const leadWkgs = filtered.filter(x => !x.watching && x.state.draft === 0 && x.gap <= 0).map(x => x.state.power / x.athlete.weight);
    const avgLeadWkg = leadWkgs.length.length ? leadWkgs.reduce((a, b) => a + b, 0) / leadWkgs.length : 0;
    const groupWkgWithoutDraft = filtered.map(x => (x.state.power + x.state.draft) / x.athlete.weight);
    const watchedActualWkg = watching.state.power / watching.athlete.weight;
    const groupMaxDraft = filtered.reduce((a, x) => Math.max(a, x.state.draft), 0);
    // const watchedMinWkg = (watching.state.power + watching.state.draft) / watching.athlete.weight - groupMaxDraft;
    state.inGroup = filtered.length > 1;
    state.watchedActualWkg = watchedActualWkg;
    state.avgLeadWkg = avgLeadWkg;
    // state.watchedMinWkg = watchedMinWkg;
    if (!state.inGroup) {
      state.groupAvgWkg = 0;
      state.watchedProjectedWkg = 0;
      state.watchedMinWkg = 0;
    } else {
      const groupAvgWkg = groupWkgWithoutDraft.reduce((a, b) => a + b, 0) / groupWkgWithoutDraft.length;
      const watchedProjectedWkg = ((groupAvgWkg * watching.athlete.weight) - watching.state.draft) / watching.athlete.weight;
      const watchedMinWkg = ((groupAvgWkg * watching.athlete.weight) - groupMaxDraft) / watching.athlete.weight;
      if (avgLeadWkg === 0) {
        state.avgLeadWkg = groupAvgWkg;
      }
      state.groupAvgWkg = groupAvgWkg;
      state.watchedProjectedWkg = watchedProjectedWkg;
      state.watchedMinWkg = watchedMinWkg;
    }
    renderer.render();
  });

  renderer.render();
}

export async function main() {
  try {
    await init();
  } catch (e) {
    console.error('ZwifTools Mod : Group Pace : Error', e);
    if (common.showError) {
      common.showError(e);
    }
  }
}

export async function settingsMain() {
  common.initInteractionListeners();
  const settingsStore = new common.SettingsStore('zwiftools-group-power-v1');
  Array.from(document.querySelectorAll('form [name="dataSmoothing"] [data-period="short"]'))
      .map(x => x.disabled = true);
  await common.initSettingsForm('form', {store: settingsStore})();
}
